/**************************************************************************************************
MiniSat -- Copyright (c) 2005, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "solver.h"

//=================================================================================================
// Simple (var/literal) helpers:

static inline int  var(lit l)  { return l >> 1; }
static inline int  sign(lit l) { return (l & 1); }

//=================================================================================================
// Debug:

//#define VERBOSEDEBUG

// For derivation output (verbosity level 2)
#define L_IND    "%-*d"
#define L_ind    decisionlevel(s)*3+3,decisionlevel(s)
#define L_LIT    "%sx%d"
#define L_lit(p) sign(p)?"~":"", (var(p))

// Just like 'assert()' but expression will be evaluated in the release version as well.
static inline void check(int expr) { assert(expr); }

static void printlits(lit* begin, lit* end)
{
    for (int i = 0; i < end - begin; i++)
        printf(L_LIT" ",L_lit(begin[i]));
}

//=================================================================================================
// Random numbers:


// Returns a random float 0 <= x < 1. Seed must never be 0.
static inline double drand(double* seed) {
    *seed *= 1389796;
    int q = (int)(*seed / 2147483647);
    *seed -= (double)q * 2147483647;
    return *seed / 2147483647; }


// Returns a random integer 0 <= x < size. Seed must never be 0.
static inline int irand(double* seed, int size) {
    return (int)(drand(seed) * size); }


//=================================================================================================
// Predeclarations:

void sort(void** array, int size, int(*comp)(const void *, const void *));

//=================================================================================================
// Clause datatype + minor functions:

struct clause_t
{
    int size_learnt;
    lit lits[0];
};

static inline int   clause_size       (clause* c)          { return c->size_learnt >> 1; }
static inline lit*  clause_begin      (clause* c)          { return c->lits; }
static inline int   clause_learnt     (clause* c)          { return c->size_learnt & 1; }
static inline float clause_activity   (clause* c)          { return *((float*)&c->lits[c->size_learnt>>1]); }
static inline void  clause_setactivity(clause* c, float a) { *((float*)&c->lits[c->size_learnt>>1]) = a; }

//=================================================================================================
// Encode literals in clause pointers:

clause* makelit (lit l)     { return (clause*)(l + l + 1); }
bool    islit   (clause* c) { return ((unsigned int)c & 1); }
lit     getlit  (clause* c) { return (lit)((unsigned int)c >> 1); }

//=================================================================================================
// Simple helpers:

static inline int     decisionlevel(solver* s)    { return vec_size(&s->trail_lim); }
static inline vec*    wlist     (solver* s, lit l){ return &s->wlists[l]; }
static inline void    vec_remove(vec* v, void* e)
{
    void** ws = vec_begin(v);
    int    j  = 0;

    for (; ws[j] != e  ; j++);
    assert(j < vec_size(v));
    for (; j < vec_size(v)-1; j++) ws[j] = ws[j+1];
    vec_resize(v,vec_size(v)-1);
}

//=================================================================================================
// Variable order functions:

static inline void updateorder(solver* s, int v)
{
    assert(s->orderpos[v] != -1);

    int*    orderpos = s->orderpos;
    double* activity = s->activity;
    int*    heap     = (int*)vec_begin(&s->order);
    int     i        = orderpos[v];
    int     x        = heap[i];
    int     parent   = (i - 1) / 2;

    while (i != 0 && activity[x] > activity[heap[parent]]){
        heap[i]           = heap[parent];
        orderpos[heap[i]] = i;
        i                 = parent;
        parent            = (i - 1) / 2;
    }
    heap[i]     = x;
    orderpos[x] = i;
}

static inline void undoorder(solver* s, int v)
{
    int* orderpos = s->orderpos;
    if (orderpos[v] == -1){
        orderpos[v] = vec_size(&s->order);
        vec_push(&s->order,(void*)v);
        updateorder(s,v);
    }
}

static int  selectvar(solver* s, float random_var_freq)
{
    lbool* values = s->assigns;

    // Random decision:
    if (drand(&s->random_seed) < random_var_freq && vec_size(&s->order) > 0){
        int next = irand(&s->random_seed,s->size);
        assert(next >= 0 && next < s->size);
        if (values[next] == l_Undef)
            return next;
    }

    // Activity based decision:

    int*    heap     = (int*)vec_begin(&s->order);
    double* activity = s->activity;
    int*    orderpos = s->orderpos;


    while (vec_size(&s->order) > 0){
        int    next  = heap[0];
        int    size  = vec_size(&s->order)-1;
        int    x     = heap[size];

        vec_resize(&s->order,size);

        orderpos[next] = -1;

        if (size > 0){
            double act   = activity[x];

            int    i     = 0;
            int    child = 1;


            while (child < size){
                if (child+1 < size && activity[heap[child]] < activity[heap[child+1]])
                    child++;

                assert(child < size);

                if (act >= activity[heap[child]])
                    break;

                heap[i]           = heap[child];
                orderpos[heap[i]] = i;
                i                 = child;
                child             = 2 * child + 1;
            }
            heap[i]           = x;
            orderpos[heap[i]] = i;
        }

        if (values[next] == l_Undef)
            return next;
    }

    return var_Undef;
}

//=================================================================================================
// Activity functions:

static inline void varrescale(solver* s) {
    double* activity = s->activity;
    for (int i = 0; i < s->size; i++)
        activity[i] *= 1e-100;
    s->var_inc *= 1e-100;
}

static inline void varbump(solver* s, int v) {
    double* activity = s->activity;
    if ((activity[v] += s->var_inc) > 1e100)
        varrescale(s);

    //printf("bump %d %f\n", v-1, activity[v]);

    if (s->orderpos[v] != -1)
        updateorder(s,v);

}

static inline void vardecay(solver* s) { s->var_inc *= s->var_decay; }

static inline void clauserescale(solver* s) {
    clause** cs = (clause**)vec_begin(&s->learnts);
    for (int i = 0; i < vec_size(&s->learnts); i++){
        float a = clause_activity(cs[i]);
        clause_setactivity(cs[i], a * 1e-20);
    }
    s->cla_inc *= 1e-20;
}


static inline void clausebump(solver* s, clause *c) {
    float a = clause_activity(c) + s->cla_inc;
    clause_setactivity(c,a);
    if (a > 1e20) clauserescale(s);
}

static inline void clausedecay(solver* s) { s->cla_inc *= s->cla_decay; }


//=================================================================================================
// Clause functions:

/* pre: size > 1 && no variable occurs twice
 */
static clause* clause_new(solver* s, lit* begin, lit* end, int learnt)
{
    assert(end - begin > 1);
    assert(learnt >= 0 && learnt < 2);
    int     size   = end - begin;
    clause* c      = (clause*)malloc(sizeof(clause) + sizeof(lit) * size + learnt * sizeof(float));
    c->size_learnt = (size << 1) | learnt;
    assert(((unsigned int)c & 1) == 0);

    for (int i = 0; i < size; i++)
        c->lits[i] = begin[i];

    if (learnt)
        *((float*)&c->lits[size]) = 0.0;

    assert(begin[0] >= 0);
    assert(begin[0] < s->size*2);
    assert(begin[1] >= 0);
    assert(begin[1] < s->size*2);

    assert(neg(begin[0]) < s->size*2);
    assert(neg(begin[1]) < s->size*2);

    //vec_push(wlist(s,neg(begin[0])),(void*)c);
    //vec_push(wlist(s,neg(begin[1])),(void*)c);

    vec_push(wlist(s,neg(begin[0])),(void*)(size > 2 ? c : makelit(begin[1])));
    vec_push(wlist(s,neg(begin[1])),(void*)(size > 2 ? c : makelit(begin[0])));

    return c;
}


static void removeclause(solver* s, clause* c)
{
    lit* lits = clause_begin(c);
    assert(neg(lits[0]) < s->size*2);
    assert(neg(lits[1]) < s->size*2);

    //vec_remove(wlist(s,neg(lits[0])),(void*)c);
    //vec_remove(wlist(s,neg(lits[1])),(void*)c);

    assert(lits[0] < s->size*2);
    vec_remove(wlist(s,neg(lits[0])),(void*)(clause_size(c) > 2 ? c : makelit(lits[1])));
    vec_remove(wlist(s,neg(lits[1])),(void*)(clause_size(c) > 2 ? c : makelit(lits[0])));

    if (clause_learnt(c)){
        s->stats.learnts--;
        s->stats.learnts_literals -= clause_size(c);
    }else{
        s->stats.clauses--;
        s->stats.clauses_literals -= clause_size(c);
    }

    free(c);
}


static lbool simplifyclause(solver* s, clause* c)
{
    assert(decisionlevel(s) == 0);
    lit*   lits   = clause_begin(c);
    lbool* values = s->assigns;

    for (int i = 0; i < clause_size(c); i++){
        lbool sig = !sign(lits[i]); sig += sig - 1;
        if (values[var(lits[i])] == sig)
            return l_True;
    }
    return l_False;
}

//=================================================================================================
// Minor (solver) functions:

static void setnvars(solver* s,int n)
{
    if (s->cap < n){

        while (s->cap < n) s->cap = s->cap*2+1;

        s->wlists    = (vec*)    realloc(s->wlists,   sizeof(vec)*s->cap*2);
        s->activity  = (double*) realloc(s->activity, sizeof(double)*s->cap);
        s->assigns   = (lbool*)  realloc(s->assigns,  sizeof(lbool)*s->cap);
        s->orderpos  = (int*)    realloc(s->orderpos, sizeof(int)*s->cap);
        s->reasons   = (clause**)realloc(s->reasons,  sizeof(clause*)*s->cap);
        s->levels    = (int*)    realloc(s->levels,   sizeof(int)*s->cap);
        s->tags      = (lbool*)  realloc(s->tags,     sizeof(lbool)*s->cap);
        s->trail     = (lit*)    realloc(s->trail,    sizeof(lit)*s->cap);
    }

    for (int var = s->size; var < n; var++){
        vec_new(&s->wlists[2*var]);
        vec_new(&s->wlists[2*var+1]);
        s->activity [var] = 0;
        s->assigns  [var] = l_Undef;
        s->orderpos [var] = var;
        s->reasons  [var] = (clause*)0;
        s->levels   [var] = 0;
        s->tags     [var] = l_Undef;

        assert(vec_size(&s->order) == var);
        vec_push(&s->order,(void*)var);
        updateorder(s,var);
    }

    s->size = n > s->size ? n : s->size;
}


static inline bool enqueue(solver* s, lit l, clause* from)
{
    lbool* values = s->assigns;
    int    v      = var(l);
    lbool  val    = values[v];
#ifdef VERBOSEDEBUG
    printf(L_IND"enqueue("L_LIT")\n", L_ind, L_lit(l));
#endif

    lbool sig = !sign(l); sig += sig - 1;
    if (val != l_Undef){
        return val == sig;
    }else{
        // New fact -- store it.
#ifdef VERBOSEDEBUG
        printf(L_IND"bind("L_LIT")\n", L_ind, L_lit(l));
#endif
        int*     levels  = s->levels;
        clause** reasons = s->reasons;

        values [v] = sig;
        levels [v] = decisionlevel(s);
        reasons[v] = from;
        s->trail[s->qtail++] = l;
        return true;
    }
}


static inline void assume(solver* s, lit l){
    assert(s->qtail == s->qhead);
    assert(s->assigns[var(l)] == l_Undef);
#ifdef VERBOSEDEBUG
    printf(L_IND"assume("L_LIT")\n", L_ind, L_lit(l));
#endif
    vec_push(&s->trail_lim,(void*)s->qtail);
    enqueue(s,l,(clause*)0);
}


static inline void canceluntil(solver* s, int level) {
    if (decisionlevel(s) <= level)
        return;

    lit*     trail   = s->trail;
    lbool*   values  = s->assigns;
    clause** reasons = s->reasons;
    int      bound   = ((int*)vec_begin(&s->trail_lim))[level];

    for (int c = s->qtail-1; c >= bound; c--) {
        int     x  = var(trail[c]);
        values [x] = l_Undef;
        reasons[x] = (clause*)0;
    }

    for (int c = s->qhead-1; c >= bound; c--)
        undoorder(s,var(trail[c]));

    s->qhead = s->qtail = bound;
    vec_resize(&s->trail_lim,level);
}

static void record(solver* s, vec* cls)
{
    assert(vec_size(cls) > 0);
    lit*    begin = (lit*)vec_begin(cls);
    lit*    end   = begin + vec_size(cls);
    clause* c     = (vec_size(cls) > 1) ? clause_new(s,begin,end,1) : (clause*)0;
    enqueue(s,*begin,c);

    if (c != 0) {
        vec_push(&s->learnts,(void*)c);
        clausebump(s,c);
        s->stats.learnts++;
        s->stats.learnts_literals += vec_size(cls);
    }
}


static double progress(solver* s)
{
    lbool*  values = s->assigns;
    int*    levels = s->levels;

    double  progress = 0;
    double  F        = 1.0 / s->size;
    for (int i = 0; i < s->size; i++)
        if (values[i] != l_Undef)
            progress += pow(F, levels[i]);
    return progress / s->size;
}

//=================================================================================================
// Major methods:

static bool removable(solver* s, lit l, int minl)
{
    lbool*   tags    = s->tags;
    clause** reasons = s->reasons;
    int*     levels  = s->levels;
    int      top     = vec_size(&s->tagged);

    assert(var(l) >= 0 && var(l) < s->size);
    assert(reasons[var(l)] != 0);
    vec_resize(&s->stack,0);
    vec_push(&s->stack,(void*)var(l));

    while (vec_size(&s->stack) > 0){
        int v = (int)vec_begin(&s->stack)[vec_size(&s->stack)-1];
        assert(v >= 0 && v < s->size);
        vec_resize(&s->stack,vec_size(&s->stack)-1);
        assert(reasons[v] != 0);
        clause* c    = reasons[v];

        if (islit(c)){
            int v = var(getlit(c));
            if (tags[v] == l_Undef && levels[v] != 0){
                if (reasons[v] != 0 && ((1 << (levels[v] & 31)) & minl)){
                    vec_push(&s->stack,(void*)v);
                    tags[v] = l_True;
                    vec_push(&s->tagged,(void*)v);
                }else{
                    int* tagged = (int*)vec_begin(&s->tagged);
                    for (int j = top; j < vec_size(&s->tagged); j++)
                        tags[tagged[j]] = l_Undef;
                    vec_resize(&s->tagged,top);
                    return false;
                }
            }
        }else{
            lit*    lits = clause_begin(c);

            for (int i = 1; i < clause_size(c); i++){
                int v = var(lits[i]);
                if (tags[v] == l_Undef && levels[v] != 0){
                    if (reasons[v] != 0 && ((1 << (levels[v] & 31)) & minl)){

                        vec_push(&s->stack,(void*)var(lits[i]));
                        tags[v] = l_True;
                        vec_push(&s->tagged,(void*)v);
                    }else{
                        int* tagged = (int*)vec_begin(&s->tagged);
                        for (int j = top; j < vec_size(&s->tagged); j++)
                            tags[tagged[j]] = l_Undef;
                        vec_resize(&s->tagged,top);
                        return false;
                    }
                }
            }
        }
    }

    return true;
}

static void analyze(solver* s, clause* c, vec* learnt)
{
    lit*     trail   = s->trail;
    lbool*   tags    = s->tags;
    clause** reasons = s->reasons;
    int*     levels  = s->levels;
    int      cnt     = 0;
    lit      p       = lit_Undef;
    int      ind     = s->qtail-1;

    vec_push(learnt,(void*)lit_Undef);

    do{
        assert(c != 0);

        if (islit(c)){
            lit q = getlit(c);
            assert(var(q) >= 0 && var(q) < s->size);
            if (tags[var(q)] == l_Undef && levels[var(q)] > 0){
                tags[var(q)] = l_True;
                vec_push(&s->tagged,(void*)var(q));
                varbump(s,var(q));
                if (levels[var(q)] == decisionlevel(s))
                    cnt++;
                else
                    vec_push(learnt,(void*)q);
            }
        }else{
            if (clause_learnt(c))
                clausebump(s,c);

            lit* lits = clause_begin(c);
            //printlits(lits,lits+clause_size(c)); printf("\n");
            for (int j = (p == lit_Undef ? 0 : 1); j < clause_size(c); j++){
                lit q = lits[j];
                assert(var(q) >= 0 && var(q) < s->size);
                if (tags[var(q)] == l_Undef && levels[var(q)] > 0){
                    tags[var(q)] = l_True;
                    vec_push(&s->tagged,(void*)var(q));
                    varbump(s,var(q));
                    if (levels[var(q)] == decisionlevel(s))
                        cnt++;
                    else
                        vec_push(learnt,(void*)q);
                }
            }
        }

        while (tags[var(trail[ind--])] == l_Undef);

        p = trail[ind+1];
        c = reasons[var(p)];
        cnt--;

    }while (cnt > 0);

    *(lit*)vec_begin(learnt) = neg(p);

    lit* lits = (lit*)vec_begin(learnt);
    int  i,j;
    int  minl = 0;
    for (i = 1; i < vec_size(learnt); i++){
        int lev = levels[var(lits[i])];
        minl    |= 1 << (lev & 31);
    }

    // simplify (full)
    for (i = j = 1; i < vec_size(learnt); i++){
        if (reasons[var(lits[i])] == 0 || !removable(s,lits[i],minl))
            lits[j++] = lits[i];
    }

    // update size of learnt + statistics
    s->stats.max_literals += vec_size(learnt);
    vec_resize(learnt,j);
    s->stats.tot_literals += j;

    // clear tags
    int* tagged = (int*)vec_begin(&s->tagged);
    for (int i = 0; i < vec_size(&s->tagged); i++)
        tags[tagged[i]] = l_Undef;
    vec_resize(&s->tagged,0);

#ifdef DEBUG
    for (int i = 0; i < s->size; i++)
        assert(tags[i] == l_Undef);
#endif

#ifdef VERBOSEDEBUG
    printf(L_IND"Learnt {", L_ind);
    for (int i = 0; i < vec_size(learnt); i++) printf(" "L_LIT, L_lit(lits[i]));
#endif
    if (vec_size(learnt) > 1){

        int max_i = 1;
        int max   = levels[var(lits[1])];
        for (int i = 2; i < vec_size(learnt); i++)
            if (levels[var(lits[i])] > max){
                max   = levels[var(lits[i])];
                max_i = i;
            }

        lit tmp     = lits[1];
        lits[1]     = lits[max_i];
        lits[max_i] = tmp;
    }
#ifdef VERBOSEDEBUG
    int lev = vec_size(learnt) > 1 ? levels[var(lits[1])] : 0;
    printf(" } at level %d\n", lev);
#endif
}


clause* propagate(solver* s)
{
    lbool*  values = s->assigns;
    clause* confl  = (clause*)0;
    lit*    lits;

    //printf("propagate\n");
    while (confl == 0 && s->qtail - s->qhead > 0){
        s->stats.propagations++;
        s->simpdb_props--;
        lit  p  = s->trail[s->qhead++];
        vec* ws = wlist(s,p);
        clause **begin = (clause**)vec_begin(ws);
        clause **end   = begin + vec_size(ws);
        clause **i, **j;
        //printf("checking lit %d: "L_LIT"\n", vec_size(ws), L_lit(p));
        for (i = j = begin; i < end; ){
            if (islit(*i)){
                *j++ = *i;
                if (!enqueue(s,getlit(*i),makelit(p))){
                    confl = s->binary;
                    (clause_begin(confl))[1] = neg(p);
                    (clause_begin(confl))[0] = getlit(*i++);

                    // Copy the remaining watches:
                    while (i < end)
                        *j++ = *i++;
                }
            }else{
                lits = clause_begin(*i);

                // Make sure the false literal is data[1]:
                lit false_lit = neg(p);
                if (lits[0] == false_lit){
                    lits[0] = lits[1];
                    lits[1] = false_lit;
                }
                assert(lits[1] == false_lit);
                //printf("checking clause: "); printlits(lits, lits+clause_size(*i)); printf("\n");

                // If 0th watch is true, then clause is already satisfied.
                lbool sig = !sign(lits[0]); sig += sig - 1;
                if (values[var(lits[0])] == sig){
                    *j++ = *i;
                }else{
                    // Look for new watch:
                    lit* stop = lits + clause_size(*i);
                    for (lit* k = lits + 2; k < stop; k++){
                        lbool sig = sign(*k); sig += sig - 1;
                        if (values[var(*k)] != sig){
                            lits[1] = *k;
                            *k = false_lit;
                            vec_push(wlist(s,neg(lits[1])),*i);
                            goto next; }
                    }

                    *j++ = *i;
                    // Clause is unit under assignment:
                    if (!enqueue(s,lits[0], *i)){
                        confl = *i++;
                        // Copy the remaining watches:
                        while (i < end)
                            *j++ = *i++;
                    }
                }
            }
        next:
            i++;
        }

        s->stats.inspects += j - (clause**)vec_begin(ws);
        vec_resize(ws,j - (clause**)vec_begin(ws));
    }

    return confl;
}

static inline int clause_cmp (const void* x, const void* y) {
    return clause_size((clause*)x) > 2 && (clause_size((clause*)y) == 2 || clause_activity((clause*)x) < clause_activity((clause*)y)) ? -1 : 1; }

void reducedb(solver* s)
{
    int      i, j;
    double   extra_lim = s->cla_inc / vec_size(&s->learnts); // Remove any clause below this activity
    clause** learnts = (clause**)vec_begin(&s->learnts);
    clause** reasons = s->reasons;

    sort(vec_begin(&s->learnts), vec_size(&s->learnts), &clause_cmp);

    for (i = j = 0; i < vec_size(&s->learnts) / 2; i++){
        if (clause_size(learnts[i]) > 2 && reasons[var(*clause_begin(learnts[i]))] != learnts[i])
            removeclause(s,learnts[i]);
        else
            learnts[j++] = learnts[i];
    }
    for (; i < vec_size(&s->learnts); i++){
        if (clause_size(learnts[i]) > 2 && reasons[var(*clause_begin(learnts[i]))] != learnts[i] && clause_activity(learnts[i]) < extra_lim)
            removeclause(s,learnts[i]);
        else
            learnts[j++] = learnts[i];
    }

    //printf("reduceDB deleted %d\n", vec_size(&s->learnts) - j);


    vec_resize(&s->learnts,j);
}

static lbool search(solver* s, int nof_conflicts, int nof_learnts)
{
    assert(s->root_level == decisionlevel(s));

    int*    levels          = s->levels;

    double  var_decay       = 0.95;
    double  clause_decay    = 0.999;
    double  random_var_freq = 0.02;

    s->stats.starts++;
    int conflictC = 0;
    s->var_decay = 1 / var_decay;
    s->cla_decay = 1 / clause_decay;
    vec_resize(&s->model,0);
    vec learnt_clause;
    vec_new(&learnt_clause);

    for (;;){
        clause* confl = propagate(s);
        if (confl != 0){
            // CONFLICT

#ifdef VERBOSEDEBUG
            printf(L_IND"**CONFLICT**\n", L_ind);
#endif
            s->stats.conflicts++; conflictC++;
            if (decisionlevel(s) == s->root_level){
                vec_delete(&learnt_clause);
                return l_False;
            }

            vec_resize(&learnt_clause,0);
            analyze(s, confl, &learnt_clause);
            int blevel = vec_size(&learnt_clause) > 1 ? levels[var(((lit*)vec_begin(&learnt_clause))[1])] : s->root_level;
            canceluntil(s,blevel);
            record(s,&learnt_clause);
            vardecay(s);
            clausedecay(s);

        }else{
            // NO CONFLICT

            if (nof_conflicts >= 0 && conflictC >= nof_conflicts){
                // Reached bound on number of conflicts:
                s->progress_estimate = progress(s);
                canceluntil(s,s->root_level);
                vec_delete(&learnt_clause);
                return l_Undef; }

            if (decisionlevel(s) == 0)
                // Simplify the set of problem clauses:
                solver_simplify(s);

            if (nof_learnts >= 0 && vec_size(&s->learnts) - s->qtail >= nof_learnts)
                // Reduce the set of learnt clauses:
                reducedb(s);

            // New variable decision:
            s->stats.decisions++;
            int next = selectvar(s,random_var_freq);

            if (next == var_Undef){
                // Model found:
                lbool* values = s->assigns;
                for (int i = 0; i < s->size; i++) vec_push(&s->model,(void*)((int)values[i]));
                canceluntil(s,s->root_level);
                vec_delete(&learnt_clause);
                return l_True;
            }

            assume(s,neg(toLit(next)));
        }
    }

    return l_Undef; // cannot happen
}

//=================================================================================================
// External solver functions:

solver* solver_new(void)
{
    solver* s = (solver*)malloc(sizeof(solver));

    // initialize vectors
    vec_new(&s->clauses);
    vec_new(&s->learnts);
    vec_new(&s->order);
    vec_new(&s->trail_lim);
    vec_new(&s->tagged);
    vec_new(&s->stack);
    vec_new(&s->model);

    // initialize arrays
    s->wlists    = 0;
    s->activity  = 0;
    s->assigns   = 0;
    s->orderpos  = 0;
    s->reasons   = 0;
    s->levels    = 0;
    s->tags      = 0;
    s->trail     = 0;


    // initialize other vars
    s->size                   = 0;
    s->cap                    = 0;
    s->qhead                  = 0;
    s->qtail                  = 0;
    s->cla_inc                = 1;
    s->cla_decay              = 1;
    s->var_inc                = 1;
    s->var_decay              = 1;
    s->root_level             = 0;
    s->simpdb_assigns         = 0;
    s->simpdb_props           = 0;
    s->random_seed            = 91648253;
    s->progress_estimate      = 0;
    s->binary                 = (clause*)malloc(sizeof(clause) + sizeof(lit)*2);
    s->binary->size_learnt    = (2 << 1);
    s->verbosity              = 0;
    s->stats.starts           = 0;
    s->stats.decisions        = 0;
    s->stats.propagations     = 0;
    s->stats.inspects         = 0;
    s->stats.conflicts        = 0;
    s->stats.clauses          = 0;
    s->stats.clauses_literals = 0;
    s->stats.learnts          = 0;
    s->stats.learnts_literals = 0;
    s->stats.max_literals     = 0;
    s->stats.tot_literals     = 0;

    return s;
}


void solver_delete(solver* s)
{
    for (int i = 0; i < vec_size(&s->clauses); i++)
        free(vec_begin(&s->clauses)[i]);

    for (int i = 0; i < vec_size(&s->learnts); i++)
        free(vec_begin(&s->learnts)[i]);

    // delete vectors
    vec_delete(&s->clauses);
    vec_delete(&s->learnts);
    vec_delete(&s->order);
    vec_delete(&s->trail_lim);
    vec_delete(&s->tagged);
    vec_delete(&s->stack);
    vec_delete(&s->model);
    free(s->binary);

    // delete arrays
    if (s->wlists != 0){
        for (int i = 0; i < s->size*2; i++)
            vec_delete(&s->wlists[i]);

        // if one is different from null, all are
        free(s->wlists);
        free(s->activity );
        free(s->assigns  );
        free(s->orderpos );
        free(s->reasons  );
        free(s->levels   );
        free(s->trail    );
        free(s->tags     );
    }

    free(s);
}


bool solver_addclause(solver* s, lit* begin, lit* end)
{
    if (begin == end) return false;

    lit *i,*j;

    //printlits(begin,end); printf("\n");
    // insertion sort
    int maxvar = var(*begin);
    for (i = begin + 1; i < end; i++){
        lit l = *i;
        maxvar = var(l) > maxvar ? var(l) : maxvar;
        for (j = i; j > begin && *(j-1) > l; j--)
            *j = *(j-1);
        *j = l;
    }
    setnvars(s,maxvar+1);

    //printlits(begin,end); printf("\n");
    lbool* values = s->assigns;

    // delete duplicates
    lit last = lit_Undef;
    for (i = j = begin; i < end; i++){
        //printf("lit: "L_LIT", value = %d\n", L_lit(*i), (sign(*i) ? -values[var(*i)] : values[var(*i)]));
        lbool sig = !sign(*i); sig += sig - 1;
        if (*i == neg(last) || sig == values[var(*i)])
            return true;   // tautology
        else if (*i != last && values[var(*i)] == l_Undef)
            last = *j++ = *i;
    }

    //printf("final: "); printlits(begin,j); printf("\n");

    if (j == begin)          // empty clause
        return false;
    else if (j - begin == 1) // unit clause
        return enqueue(s,*begin,(clause*)0);

    // create new clause
    vec_push(&s->clauses,clause_new(s,begin,j,0));


    s->stats.clauses++;
    s->stats.clauses_literals += j - begin;

    return true;
}


bool   solver_simplify(solver* s)
{
    assert(decisionlevel(s) == 0);

    if (propagate(s) != 0)
        return false;

    if (s->qhead == s->simpdb_assigns || s->simpdb_props > 0)
        return true;

    clause** reasons = s->reasons;

    for (int type = 0; type < 2; type++){
        vec*     cs  = type ? &s->learnts : &s->clauses;
        clause** cls = (clause**)vec_begin(cs);

        int     j = 0;
        for (int i = 0; i < vec_size(cs); i++){
            if (reasons[var(*clause_begin(cls[i]))] != cls[i] &&
                simplifyclause(s,cls[i]) == l_True)
                removeclause(s,cls[i]);
            else
                cls[j++] = cls[i];
        }
        vec_resize(cs,j);
    }

    s->simpdb_assigns = s->qhead;
    // (shouldn't depend on 'stats' really, but it will do for now)
    s->simpdb_props   = s->stats.clauses_literals + s->stats.learnts_literals;

    return true;
}


bool   solver_solve(solver* s, lit* begin, lit* end)
{
    double  nof_conflicts = 100;
    double  nof_learnts   = solver_nclauses(s) / 3;
    lbool   status        = l_Undef;
    lbool*  values        = s->assigns;

    for (lit* i = begin; i < end; i++)
        if ((sign(*i) ? -values[var(*i)] : values[var(*i)]) == l_False || (assume(s,*i), propagate(s) != 0)){
            canceluntil(s,0);
            return false; }

    s->root_level = decisionlevel(s);

    if (s->verbosity >= 1){
        printf("==================================[MINISAT]===================================\n");
        printf("| Conflicts |     ORIGINAL     |              LEARNT              | Progress |\n");
        printf("|           | Clauses Literals |   Limit Clauses Literals  Lit/Cl |          |\n");
        printf("==============================================================================\n");
    }

    while (status == l_Undef){
        if (s->verbosity >= 1){
            printf("| %9d | %7d %8d | %7d %7d %8d %7.1f | %6.3f %% |\n", (int)s->stats.conflicts,(int)s->stats.clauses, (int)s->stats.clauses_literals,(int)nof_learnts, (int)s->stats.learnts, (int)s->stats.learnts_literals,(double)s->stats.learnts_literals / (double)s->stats.learnts,s->progress_estimate*100);
            fflush(stdout);
        }
        status = search(s,(int)nof_conflicts, (int)nof_learnts);
        nof_conflicts *= 1.5;
        nof_learnts   *= 1.1;
    }
    if (s->verbosity >= 1)
        printf("==============================================================================)\n");

    canceluntil(s,0);
    return status != l_False;
}


int solver_nvars(solver* s)
{
    return s->size;
}


int solver_nclauses(solver* s)
{
    return vec_size(&s->clauses);
}

//=================================================================================================
// Sorting functions (sigh):

static inline void selectionsort(void** array, int size, int(*comp)(const void *, const void *))
{
    int     i, j, best_i;
    void*   tmp;

    for (i = 0; i < size-1; i++){
        best_i = i;
        for (j = i+1; j < size; j++){
            if (comp(array[j], array[best_i]) < 0)
                best_i = j;
        }
        tmp = array[i]; array[i] = array[best_i]; array[best_i] = tmp;
    }
}


static void sortrnd(void** array, int size, int(*comp)(const void *, const void *), double* seed)
{
    if (size <= 15)
        selectionsort(array, size, comp);

    else{
        void*       pivot = array[irand(seed, size)];
        void*       tmp;
        int         i = -1;
        int         j = size;

        for(;;){
            do i++; while(comp(array[i], pivot)<0);
            do j--; while(comp(pivot, array[j])<0);

            if (i >= j) break;

            tmp = array[i]; array[i] = array[j]; array[j] = tmp;
        }

        sortrnd(array    , i     , comp, seed);
        sortrnd(&array[i], size-i, comp, seed);
    }
}

void sort(void** array, int size, int(*comp)(const void *, const void *))
{
    double seed = 91648253;
    sortrnd(array,size,comp,&seed);
}
